"""
 mod_ns_recs.py : Module, checks for NS entries

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

import mod_get_dnservs
import mod_soa_recs

class Module(TestModule):
    """Gets NS records

       The module uses data resolved by the mod_get_dnservs and mod_soa_recs
       module. Module creates for each DNS server IP a resolver. Then asks
       by using theresolvers for NS entries.

       The obtained NS records are stored in a tree structure created
       from python dictionaries. This structure is copied into a global
       environment variable, on order to keep the results accessible for
       other modules.

       The module checks whether all DNS machines, which are listed from the
       result of module mod_get_dnservs, are also listed on other servers.

       Be careful when creating IPv6 based resolvers. This may cause
       the application to crash on no IPv6 networks. Be also careful when
       trying to disable the IPv4 functionality.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Get NS records")
        self.dns_env = self.get_module_modenv(mod_get_dnservs)
        self.soa_env = self.get_module_modenv(mod_soa_recs)

    def execute(self):

        res = {}

        res["dnservs"] = {}
        domain = self.ldns.ldns_dname_new_frm_str(self.dns_env["domain"])
        resolver = self.ldns.ldns_resolver_new()

        # for each server name
        for servname in self.dns_env["dnservs"].keys():
            res["dnservs"][servname] = {}

            # for ipv4 addresses
            if (not self.args.options.noipv4
                and ("ipv4" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv4"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv4"].keys():
                    res["dnservs"][servname]["ipv4"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv4"][addr]["alive"]):
                        continue
    
                    # create resolver
                    while resolver.pop_nameserver():
                         pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv4"][addr]["rdf"])
                    if (not resolver):
                        continue
    
                    # ask for NS for every argument
                    self.verb(2, "NS records for", servname, "ipv4", addr)
                    pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_NS,
                        self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                    if (not pkt):
                        self.err(0, "Resolver response error")
                        return False
                    ns = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                        self.ldns.LDNS_SECTION_ANSWER)
                    res["dnservs"][servname]["ipv4"][addr]["NS"] = {}
    
                    if (not self.args.options.noipv4):
                        # find ipv4 dresses in glue
                        gluea = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                        # ask for each ipv4 address of all nameservers
                        for sname in self.dns_env["dnservs"].keys():
                            if (sname not in res["dnservs"][servname]["ipv4"][addr]["NS"].keys()):
                                res["dnservs"][servname]["ipv4"][addr]["NS"][sname] = {}
                            dreqname = self.ldns.ldns_dname_new_frm_str(sname)
                            subpkt = resolver.query(dreqname, self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                            if (not subpkt):
                                self.err(0, "Resolver response error")
                                return False
                            a = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_SECTION_ANSWER)
                            if (not a):
                                continue
    
                            # write down results
                            res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["A"] = []
                            for a_rr in a.rrs():
                                res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["A"].append(str(a_rr.rdf(0)))
                                if (not gluea):
                                    for ipv4 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["A"]:
                                        self.verb(2, "server:", sname, "discovered-ipv4:", ipv4)
                                    break
                                number = int(gluea.rr_count())
                                count = 0
                                for gluea_rr in gluea.rrs():
                                    if (gluea_rr.owner() == dreqname):
                                        # if found in glue, write down glue
                                        if ("glueA" not in res["dnservs"][servname]["ipv4"][addr]["NS"][sname].keys()):
                                            res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueA"] = []
                                        if (str(gluea_rr.rdf(0)) not in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueA"]):
                                            res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueA"].append(str(gluea_rr.rdf(0)))
                                            for ipv4 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["A"]:
                                                for gipv4 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueA"]:
                                                    self.verb(2, "server:", sname, "discovered-ipv4:", ipv4, "glue-ipv4:", gipv4)
                                    else:
                                        count += 1
                                    if (count == number):
                                        # if not in glue
                                        for ipv4 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["A"]:
                                            self.verb(2, "server:", sname, "discovered-ipv4:", ipv4)
                                        
                    if (self.args.options.ipv6):
                        # find ipv6 addresses in glue
                        glueaaaa = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                        # ask for each ipv6 address of all nameservers
                        for sname in self.dns_env["dnservs"].keys():
                            if (sname not in res["dnservs"][servname]["ipv4"][addr]["NS"].keys()):
                                res["dnservs"][servname]["ipv4"][addr]["NS"][sname] = {}
                            dreqname = self.ldns.ldns_dname_new_frm_str(sname)
                            subpkt = resolver.query(dreqname, self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                            if (not subpkt):
                                self.err(0, "Resolver response error")
                                return False
                            aaaa = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_SECTION_ANSWER)
                            if (not aaaa):
                                continue
    
                            # write down results
                            res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["AAAA"] = []
                            for aaaa_rr in aaaa.rrs():
                                res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["AAAA"].append(str(aaaa_rr.rdf(0)))
                                if (not glueaaaa):
                                    for ipv6 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["AAAA"]:
                                        self.verb(2, "server:", sname, "discovered-ipv6:", ipv6)
                                    break
                                number = int(glueaaaa.rr_count())
                                count = 0
                                for glueaaaa_rr in glueaaaa.rrs():
                                    if (glueaaaa_rr.owner() == dreqname):
                                        # if found in glue, write down glue
                                        if ("glueAAAA" not in res["dnservs"][servname]["ipv4"][addr]["NS"][sname].keys()):
                                            res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueAAAA"] = []
                                        if (str(glueaaaa_rr.rdf(0)) not in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueAAAA"]):
                                            res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueAAAA"].append(str(glueaaaa_rr.rdf(0)))
                                            for ipv6 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["AAAA"]:
                                                for gipv6 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["glueAAAA"]: 
                                                    self.verb(2, "server:", sname, "discovered-ipv6:", ipv6, "glue-ipv6:", gipv6)
                                    else:
                                        count += 1
                                    if (count == number):
                                        # if not in glue
                                        for ipv6 in res["dnservs"][servname]["ipv4"][addr]["NS"][sname]["AAAA"]:
                                            self.verb(2, "server:", sname, "discovered-ipv6:", ipv6)
    
                    self.verb(2, "")
                        

            # for ipv6 addresses
            if (self.args.options.resolver6
                and ("ipv6" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv6"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv6"].keys():
                    res["dnservs"][servname]["ipv6"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv6"][addr]["alive"]):
                        continue

                    # create resolver
                    while resolver.pop_nameserver():
                        pass
                    resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv6"][addr]["rdf"])
                    if (not resolver):
                        continue

                    # ask for NS for every argument
                    self.verb(2, "NS records for", servname, "ipv6", addr)
                    pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_NS,
                        self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                    if (not pkt):
                        self.err(0, "Resolver response error")
                        return False
                    ns = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                        self.ldns.LDNS_SECTION_ANSWER)
                    res["dnservs"][servname]["ipv6"][addr]["NS"] = {}

                    if (not self.args.options.noipv4):
                        # find ipv4 dresses in glue
                        gluea = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                        # ask for each ipv4 address of all nameservers
                        for sname in self.dns_env["dnservs"].keys():
                            if (sname not in res["dnservs"][servname]["ipv6"][addr]["NS"].keys()):
                                res["dnservs"][servname]["ipv6"][addr]["NS"][sname] = {}
                            dreqname = self.ldns.ldns_dname_new_frm_str(sname)
                            subpkt = resolver.query(dreqname, self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                            if (not subpkt):
                                self.err(0, "Resolver response error")
                                return False
                            a = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_SECTION_ANSWER)
                            if (not a):
                                continue

                            # write down results
                            res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["A"] = []
                            for a_rr in a.rrs():
                                res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["A"].append(str(a_rr.rdf(0)))
                                if (not gluea):
                                    for ipv4 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["A"]:
                                        self.verb(2, "server:", sname, "discovered-ipv4:", ipv4)
                                    break
                                number = int(gluea.rr_count())
                                count = 0
                                for gluea_rr in gluea.rrs():
                                    if (gluea_rr.owner() == dreqname):
                                        # if found in glue, write down glue
                                        if ("glueA" not in res["dnservs"][servname]["ipv6"][addr]["NS"][sname].keys()):
                                            res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueA"] = []
                                        if (str(gluea_rr.rdf(0)) not in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueA"]):
                                            res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueA"].append(str(gluea_rr.rdf(0)))
                                            for ipv4 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["A"]:
                                                for gipv4 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueA"]:
                                                    self.verb(2, "server:", sname, "discovered-ipv4:", ipv4, "glue-ipv4:", gipv4)
                                    else:
                                        count += 1
                                    if (count == number):
                                        # if not in glue
                                        for ipv4 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["A"]:
                                            self.verb(2, "server:", sname, "discovered-ipv4:", ipv4)

                    if (self.args.options.ipv6):
                        # find ipv6 addresses in glue
                        glueaaaa = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                            self.ldns.LDNS_SECTION_ADDITIONAL)
                        # ask for each ipv6 address of all nameservers
                        for sname in self.dns_env["dnservs"].keys():
                            if (sname not in res["dnservs"][servname]["ipv6"][addr]["NS"].keys()):
                                res["dnservs"][servname]["ipv6"][addr]["NS"][sname] = {}
                            dreqname = self.ldns.ldns_dname_new_frm_str(sname)
                            subpkt = resolver.query(dreqname, self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)
                            if (not subpkt):
                                self.err(0, "Resolver response error")
                                return False
                            aaaa = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_SECTION_ANSWER)
                            if (not aaaa):
                                continue

                            # write down results
                            res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["AAAA"] = []
                            for aaaa_rr in aaaa.rrs():
                                res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["AAAA"].append(str(aaaa_rr.rdf(0)))
                                if (not glueaaaa):
                                    for ipv6 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["AAAA"]:
                                        self.verb(2, "server:", sname, "discovered-ipv6:", ipv6)
                                    break
                                number = int(glueaaaa.rr_count())
                                count = 0
                                for glueaaaa_rr in glueaaaa.rrs():
                                    if (glueaaaa_rr.owner() == dreqname):
                                        # if found in glue, write down glue
                                        if ("glueAAAA" not in res["dnservs"][servname]["ipv6"][addr]["NS"][sname].keys()):
                                            res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueAAAA"] = []
                                        if (str(glueaaaa_rr.rdf(0)) not in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueAAAA"]):
                                            res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueAAAA"].append(str(glueaaaa_rr.rdf(0)))
                                            for ipv6 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["AAAA"]:
                                                for gipv6 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["glueAAAA"]:
                                                    self.verb(2, "server:", sname, "discovered-ipv6:", ipv6, "glue-ipv6:", gipv6)
                                    else:
                                        count += 1
                                    if (count == number):
                                        # if not in glue
                                        for ipv6 in res["dnservs"][servname]["ipv6"][addr]["NS"][sname]["AAAA"]:
                                            self.verb(2, "server:", sname, "discovered-ipv6:", ipv6)

                    self.verb(2, "")

        # check whether all nameservers contain all namervers in NS records
        serv_in_all = {}
        # cerate entries for all nameserver names
        for servname in res["dnservs"].keys():
            serv_in_all[servname] = True
        for servname in serv_in_all.keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in res["dnservs"][servname]["ipv4"].keys():
                    if ("NS" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                        # test all names in every nameserver
                        for srv in serv_in_all.keys():
                            present = srv in res["dnservs"][servname]["ipv4"][ipv4]["NS"]
                            self.verb(1, servname, ipv4, "ns entry for", srv, "present:", present)
                            serv_in_all[srv] = serv_in_all[srv] and present
                    else:
                        # no NS entries
                        self.verb(1, servname, ipv4, "no NS entries in")
                        for srv in serv_in_all.keys():
                            serv_in_all[srv] = False
            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():
                    if ("NS" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                        # test all names in every nameserver
                        for srv in serv_in_all.keys():
                            present = srv in res["dnservs"][servname]["ipv6"][ipv6]["NS"]
                            self.verb(1, servname, ipv6, "ns entry for", srv, "present:", present)
                            serv_in_all[srv] = serv_in_all[srv] and present
                    else:
                        # no NS entries
                        self.verb(1, servname, ipv6, "no NS entries in")
                        for srv in serv_in_all.keys():
                            serv_in_all[srv] = False
        self.verb(1, "")
        all_present = True
        for srv in serv_in_all.keys():
            all_present = all_present and serv_in_all[srv]
        if (all_present):
            self.verb(0, "All DNS servers listed in NS records on each server: yes")
        else:
            self.verb(0, "All DNS servers listed in NS records on every server: no")
        self.verb(0, "")

        self.set_my_modenv(res)
        return True
